"""This module contains preset settings for defining cross-validation iterators"""
from __future__ import absolute_import, print_function, division, unicode_literals


__all__ = [
    'k_fold',
    'stratified_k_fold',
    'shuffle_split',
    'stratified_shuffle_split',
    'leave_one_out',
    'leave_p_out',
    'group_k_fold',
    'time_series_split'
]

k_fold = {
    'name': 'K-fold Cross Validation',
    'source':
    """from sklearn.model_selection import KFold

def return_splits_iterable(X, y):
    \"\"\"This function returns an iterable that splits the given dataset
    K times into different train-test splits.
    \"\"\"
    RANDOM_STATE = 8
    N_SPLITS = 5
    SHUFFLE = True

    return KFold(n_splits=N_SPLITS, random_state=RANDOM_STATE, shuffle=SHUFFLE).split(X, y)
"""
}

stratified_k_fold = {
    'name': 'Stratified K-fold Cross Validation',
    'source':
    """from sklearn.model_selection import StratifiedKFold

def return_splits_iterable(X, y):
    \"\"\"This function returns an iterable that splits the given dataset
    K times into different stratified train-test splits.
    \"\"\"
    RANDOM_STATE = 8
    N_SPLITS = 5
    SHUFFLE = True

    return StratifiedKFold(n_splits=N_SPLITS, random_state=RANDOM_STATE, shuffle=SHUFFLE).split(X, y)
"""
}

shuffle_split = {
    'name': 'Shuffle Split',
    'source':
    """from sklearn.model_selection import ShuffleSplit

def return_splits_iterable(X, y):
    \"\"\"This function returns an iterable that splits the given dataset
    randomly into different train-test splits. For applications where you
    only need a single train-test split (large datasets), you can do this
    by setting `n_splits` to 1.
    \"\"\"
    RANDOM_STATE = 8
    N_SPLITS = 1
    TEST_SIZE = 0.25
    TRAIN_SIZE = None

    return ShuffleSplit(n_splits=N_SPLITS, random_state=RANDOM_STATE,
                        test_size=TEST_SIZE, train_size=TRAIN_SIZE).split(X, y)
"""
}

stratified_shuffle_split = {
    'name': 'Stratified Shuffle Split',
    'source':
    """from sklearn.model_selection import StratifiedShuffleSplit

def return_splits_iterable(X, y):
    \"\"\"This function returns an iterable that splits the given dataset
    randomly into different stratified train-test splits. For applications where you
    only need a single train-test split (large datasets), you can do this
    by setting `n_splits` to 1.
    \"\"\"
    RANDOM_STATE = 8
    N_SPLITS = 1
    TEST_SIZE = 0.25
    TRAIN_SIZE = None

    return StratifiedShuffleSplit(n_splits=N_SPLITS, random_state=RANDOM_STATE,
                                  test_size=TEST_SIZE, train_size=TRAIN_SIZE).split(X, y)
"""
}


leave_one_out = {
    'name': 'Leave One Out',
    'source':
    """from sklearn.model_selection import LeaveOneOut

def return_splits_iterable(X, y):
    \"\"\" This function returns an iterable that splits the dataset, where
    train dataset is n-1, leaving only one out to test. 
    \"\"\"

    return LeaveOneOut().split(X, y)
"""
}

leave_p_out = {
    'name': 'Leave P Out',
    'source':
    """from sklearn.model_selection import LeavePOut
    
def return_splits_iterable(X, y):
    \"\"\" This function returns an iterable that splits the dataset, where
    train dataset is n-p, leaving p out to test. 
    \"\"\"
    P=2

    return LeavePOut(p=P).split(X, y)
"""
}

group_k_fold = {
    'name': 'Group K-fold',
    'source':
    """from sklearn.model_selection import GroupKFold

def return_splits_iterable(X, y):
    \"\"\" This function returns a K-fold iterator variant
    with non-overlapping groups. The number of distict groups has to be at least 
    equal to the number of folds
    \"\"\"

    N_SPLITS = 3
    
    GROUPS = None  # You must fill this with your own group labels

    return GroupKFold(n_splits=N_SPLITS).split(X, y, groups=GROUPS)
"""
}

time_series_split = {
    'name': 'Time Series Split',
    'source':
    """from sklearn.model_selection import TimeSeriesSplit

def return_splits_iterable(X, y):
    \"\"\" This function is a variation of Kfold where it splits
    time-series dataset at fixed time intervals.
    \"\"\"

    N_SPLITS = 3

    return TimeSeriesSplit(n_splits=N_SPLITS).split(X, y)
"""
}




from __future__ import absolute_import, print_function, division, unicode_literals
import unittest
import numpy as np
from sklearn.datasets import load_iris, load_breast_cancer, load_boston
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.model_selection import cross_val_predict
from xcessiv import functions
from xcessiv.presets import metricsetting


clf = LogisticRegression(random_state=8)

multiclass_X, multiclass_y = load_iris(return_X_y=True)
multiclass_preds = cross_val_predict(clf, multiclass_X, multiclass_y, method='predict')
multiclass_probas = cross_val_predict(clf, multiclass_X, multiclass_y, method='predict_proba')

binary_X, binary_y = load_breast_cancer(return_X_y=True)
binary_preds = cross_val_predict(clf, binary_X, binary_y, method='predict')
binary_probas = cross_val_predict(clf, binary_X, binary_y, method='predict_proba')

regression_X, regression_y = load_boston(return_X_y=True)
reg = LinearRegression()
regression_preds = cross_val_predict(reg, regression_X, regression_y, method='predict')


class TestAccuracyFromScores(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.accuracy_from_scores['source'])

        assert np.round(module.metric_generator(binary_y, binary_probas), 2) == 0.95
        assert np.round(module.metric_generator(multiclass_y, multiclass_probas), 2) == 0.95

        del module


class TestAccuracyFromPreds(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.accuracy_from_preds['source'])

        assert np.round(module.metric_generator(binary_y, binary_preds), 2) == 0.95
        assert np.round(module.metric_generator(multiclass_y, multiclass_preds), 2) == 0.95

        del module


class TestRecallFromScores(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.recall_from_scores['source'])

        assert np.round(module.metric_generator(binary_y, binary_probas), 2) == 0.97
        assert np.round(module.metric_generator(multiclass_y, multiclass_probas), 2) == 0.95

        del module


class TestRecallFromPreds(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.recall_from_preds['source'])

        assert np.round(module.metric_generator(binary_y, binary_preds), 2) == 0.97
        assert np.round(module.metric_generator(multiclass_y, multiclass_preds), 2) == 0.95

        del module


class TestPrecisionFromScores(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.precision_from_scores['source'])

        assert np.round(module.metric_generator(binary_y, binary_probas), 2) == 0.95
        assert np.round(module.metric_generator(multiclass_y, multiclass_probas), 2) == 0.95

        del module


class TestPrecisionFromPreds(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.precision_from_preds['source'])

        assert np.round(module.metric_generator(binary_y, binary_preds), 2) == 0.95
        assert np.round(module.metric_generator(multiclass_y, multiclass_preds), 2) == 0.95

        del module


class TestF1ScoreFromScores(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.f1_score_from_scores['source'])

        assert np.round(module.metric_generator(binary_y, binary_probas), 2) == 0.96
        assert np.round(module.metric_generator(multiclass_y, multiclass_probas), 2) == 0.95

        del module


class TestF1ScoreFromPreds(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.f1_score_from_preds['source'])

        assert np.round(module.metric_generator(binary_y, binary_preds), 2) == 0.96
        assert np.round(module.metric_generator(multiclass_y, multiclass_preds), 2) == 0.95

        del module


class TestROCAUCFromScores(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(
            metricsetting.roc_auc_score_from_scores['source']
        )

        assert np.round(module.metric_generator(binary_y, binary_probas), 2) == 0.99
        assert np.round(module.metric_generator(multiclass_y, multiclass_probas), 2) == 0.99

        del module


class TestMAE(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.mae['source'])

        assert np.round(module.metric_generator(regression_y, regression_preds), 2) == 6.99

        del module


class TestMSE(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.mse['source'])

        assert np.round(module.metric_generator(regression_y, regression_preds), 2) == 168.09

        del module


class TestMedianAbsoluteError(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.median_absolute_error['source'])

        assert np.round(module.metric_generator(regression_y, regression_preds), 2) == 3.72

        del module


class TestR2Score(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.r2_score['source'])

        assert np.round(module.metric_generator(regression_y, regression_preds), 2) == -0.99

        del module


class TestExplainedVarianceScore(unittest.TestCase):
    def test_source(self):
        module = functions.import_string_code_as_module(metricsetting.explained_variance_score['source'])

        assert np.round(module.metric_generator(regression_y, regression_preds), 2) == -0.89

        del module



import sklearn
if sklearn.__version__.startswith('0.18'):
    from sklearn.pipeline import _BasePipeline as bp
else:
    from sklearn.utils.metaestimators import _BaseComposition as bp
import numpy as np


class XcessivStackedEnsemble(bp):
    """Contains the class for the Xcessiv stacked ensemble"""
    def __init__(self, base_learners, meta_feature_generators,
                 secondary_learner, cv_function):
        super(XcessivStackedEnsemble, self).__init__()

        self.base_learners = base_learners
        self.meta_feature_generators = meta_feature_generators
        self.secondary_learner = secondary_learner
        self.cv_function = cv_function
        self._named_learners = [('bl{}'.format(idx), base_learner) for idx, base_learner
                               in enumerate(base_learners)]
        self._named_learners.append(('secondary-learner', secondary_learner))

    def get_params(self, deep=True):
        """Get parameters for this estimator.

        Args:

        deep (boolean, optional): If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        params: mapping of string to any Parameter names mapped to their values.
        """
        return self._get_params('_named_learners', deep=deep)

    def set_params(self, **params):
        """Set the parameters of this estimator."""
        self._set_params('_named_learners', **params)
        return self

    def fit(self, X, y):
        print('Fitting {} base learners'.format(len(self.base_learners)))

        all_learner_meta_features = []
        for idx, base_learner in enumerate(self.base_learners):

            single_learner_meta_features = []
            test_indices = []
            for num, (train_idx, test_idx) in enumerate(self.cv_function(X, y)):
                print('Fold {} of base learner {}'.format(num+1, idx+1))

                base_learner.fit(X[train_idx], y[train_idx])

                preds = getattr(base_learner, self.meta_feature_generators[idx])(X[test_idx])

                if len(preds.shape) == 1:
                    preds = preds.reshape(-1, 1)

                single_learner_meta_features.append(
                    preds
                )

                test_indices.append(test_idx)

            single_learner_meta_features = np.concatenate(single_learner_meta_features)
            all_learner_meta_features.append(single_learner_meta_features)

        all_learner_meta_features = np.concatenate(all_learner_meta_features, axis=1)
        test_indices = np.concatenate(test_indices)  # reorganized order due to CV

        print('Fitting meta-learner')

        self.secondary_learner.fit(all_learner_meta_features, y[test_indices])

        return self

    def _process_using_meta_feature_generator(self, X, meta_feature_generator):
        """Process using secondary learner meta-feature generator

        Since secondary learner meta-feature generator can be anything e.g. predict, predict_proba,
        this internal method gives the ability to use any string. Just make sure secondary learner
        has the method.

        Args:
            X (array-like): Features array

            meta_feature_generator (str, unicode): Method for use by secondary learner
        """

        all_learner_meta_features = []
        for idx, base_learner in enumerate(self.base_learners):
            single_learner_meta_features = getattr(base_learner,
                                                   self.meta_feature_generators[idx])(X)

            if len(single_learner_meta_features.shape) == 1:
                single_learner_meta_features = single_learner_meta_features.reshape(-1, 1)
            all_learner_meta_features.append(single_learner_meta_features)

        all_learner_meta_features = np.concatenate(all_learner_meta_features, axis=1)

        out = getattr(self.secondary_learner, meta_feature_generator)(all_learner_meta_features)

        return out



"""This module contains functions for the automated runs"""
from __future__ import absolute_import, print_function, division, unicode_literals
from rq import get_current_job
from xcessiv import functions
from xcessiv import models
from xcessiv import constants
import numpy as np
import os
import sys
import traceback
from six import iteritems
import numbers
from bayes_opt import BayesianOptimization


def return_func_to_optimize(path, session, base_learner_origin, default_params,
                            metric_to_optimize, invert_metric, integers):
    """Creates the function to be optimized by Bayes Optimization.

    The function automatically handles the case of already existing base learners, and if
    no base learner for the hyperparameters exists yet, creates one and updates it in the
    usual way.

    Args:
        path (str): Path to Xcessiv notebook

        session: Database session passed down

        base_learner_origin: BaseLearnerOrigin object

        default_params (dict): Dictionary containing default params of estimator

        metric_to_optimize (str, unicode): String containing name of metric to optimize

        invert_metric (bool): Specifies whether metric should be inverted e.g. losses

        integers (set): Set of strings that specify which hyperparameters are integers

    Returns:
        func_to_optimize (function): Function to be optimized
    """
    def func_to_optimize(**params):
        base_estimator = base_learner_origin.return_estimator()
        base_estimator.set_params(**default_params)
        # For integer hyperparameters, make sure they are rounded off
        params = dict((key, val) if key not in integers else (key, int(val))
                      for key, val in iteritems(params))
        base_estimator.set_params(**params)
        hyperparameters = functions.make_serializable(base_estimator.get_params())

        # Look if base learner already exists
        base_learner = session.query(models.BaseLearner).\
            filter_by(base_learner_origin_id=base_learner_origin.id,
                      hyperparameters=hyperparameters).first()

        calculate_only = False

        # If base learner exists and has finished, just return its result
        if base_learner and base_learner.job_status == 'finished':
            if invert_metric:
                return -base_learner.individual_score[metric_to_optimize]
            else:
                return base_learner.individual_score[metric_to_optimize]

        # else if base learner exists but is unfinished, just calculate the result without storing
        elif base_learner and base_learner.job_status != 'finished':
            calculate_only = True

        # else if base learner does not exist, create it
        else:
            base_learner = models.BaseLearner(hyperparameters,
                                              'started',
                                              base_learner_origin)
            base_learner.job_id = get_current_job().id
            session.add(base_learner)
            session.commit()

        try:
            est = base_learner.return_estimator()
            extraction = session.query(models.Extraction).first()
            X, y = extraction.return_train_dataset()
            return_splits_iterable = functions.import_object_from_string_code(
                extraction.meta_feature_generation['source'],
                'return_splits_iterable'
            )

            meta_features_list = []
            trues_list = []
            for train_index, test_index in return_splits_iterable(X, y):
                X_train, X_test = X[train_index], X[test_index]
                y_train, y_test = y[train_index], y[test_index]
                est = est.fit(X_train, y_train)
                meta_features_list.append(
                    getattr(est, base_learner.base_learner_origin.
                            meta_feature_generator)(X_test)
                )
                trues_list.append(y_test)
            meta_features = np.concatenate(meta_features_list, axis=0)
            y_true = np.concatenate(trues_list)

            for key in base_learner.base_learner_origin.metric_generators:
                metric_generator = functions.import_object_from_string_code(
                    base_learner.base_learner_origin.metric_generators[key],
                    'metric_generator'
                )
                base_learner.individual_score[key] = metric_generator(y_true, meta_features)

            # Only do this if you want to save things
            if not calculate_only:
                meta_features_path = base_learner.meta_features_path(path)

                if not os.path.exists(os.path.dirname(meta_features_path)):
                    os.makedirs(os.path.dirname(meta_features_path))

                np.save(meta_features_path, meta_features, allow_pickle=False)
                base_learner.job_status = 'finished'
                base_learner.meta_features_exists = True
                session.add(base_learner)
                session.commit()

            if invert_metric:
                return -base_learner.individual_score[metric_to_optimize]
            else:
                return base_learner.individual_score[metric_to_optimize]

        except:
            session.rollback()
            base_learner.job_status = 'errored'
            base_learner.description['error_type'] = repr(sys.exc_info()[0])
            base_learner.description['error_value'] = repr(sys.exc_info()[1])
            base_learner.description['error_traceback'] = \
                traceback.format_exception(*sys.exc_info())
            session.add(base_learner)
            session.commit()
            raise
    return func_to_optimize


def start_naive_bayes(automated_run, session, path):
    """Starts naive bayes automated run

    Args:
        automated_run (xcessiv.models.AutomatedRun): Automated run object

        session: Valid SQLAlchemy session

        path (str, unicode): Path to project folder
    """
    module = functions.import_string_code_as_module(automated_run.source)
    random_state = 8 if not hasattr(module, 'random_state') else module.random_state
    assert module.metric_to_optimize in automated_run.base_learner_origin.metric_generators

    # get non-searchable parameters
    base_estimator = automated_run.base_learner_origin.return_estimator()
    base_estimator.set_params(**module.default_params)
    default_params = functions.make_serializable(base_estimator.get_params())
    non_searchable_params = dict((key, val) for key, val in iteritems(default_params)
                                 if key not in module.pbounds)

    # get already calculated base learners in search space
    existing_base_learners = []
    for base_learner in automated_run.base_learner_origin.base_learners:
        if not base_learner.job_status == 'finished':
            continue
        in_search_space = True
        for key, val in iteritems(non_searchable_params):
            if base_learner.hyperparameters[key] != val:
                in_search_space = False
                break  # If no match, move on to the next base learner
        if in_search_space:
            existing_base_learners.append(base_learner)

    # build initialize dictionary
    target = []
    initialization_dict = dict((key, list()) for key in module.pbounds.keys())
    for base_learner in existing_base_learners:
        # check if base learner's searchable hyperparameters are all numerical
        all_numerical = True
        for key in module.pbounds.keys():
            if not isinstance(base_learner.hyperparameters[key], numbers.Number):
                all_numerical = False
                break
        if not all_numerical:
            continue  # if there is a non-numerical hyperparameter, skip this.

        for key in module.pbounds.keys():
            initialization_dict[key].append(base_learner.hyperparameters[key])
        target.append(base_learner.individual_score[module.metric_to_optimize])
    initialization_dict['target'] = target if not module.invert_metric \
        else list(map(lambda x: -x, target))
    print('{} existing in initialization dictionary'.
          format(len(initialization_dict['target'])))

    # Create function to be optimized
    func_to_optimize = return_func_to_optimize(
        path, session, automated_run.base_learner_origin, module.default_params,
        module.metric_to_optimize, module.invert_metric, set(module.integers)
    )

    # Create Bayes object
    bo = BayesianOptimization(func_to_optimize, module.pbounds)

    bo.initialize(initialization_dict)

    np.random.seed(random_state)

    bo.maximize(**module.maximize_config)


def start_tpot(automated_run, session, path):
    """Starts a TPOT automated run that exports directly to base learner setup

    Args:
        automated_run (xcessiv.models.AutomatedRun): Automated run object

        session: Valid SQLAlchemy session

        path (str, unicode): Path to project folder
    """
    module = functions.import_string_code_as_module(automated_run.source)
    extraction = session.query(models.Extraction).first()
    X, y = extraction.return_train_dataset()

    tpot_learner =  module.tpot_learner

    tpot_learner.fit(X, y)

    temp_filename = os.path.join(path, 'tpot-temp-export-{}'.format(os.getpid()))
    tpot_learner.export(temp_filename)

    with open(temp_filename) as f:
        base_learner_source = f.read()

    base_learner_source = constants.tpot_learner_docstring + base_learner_source

    try:
        os.remove(temp_filename)
    except OSError:
        pass

    blo = models.BaseLearnerOrigin(
        source=base_learner_source,
        name='TPOT Learner',
        meta_feature_generator='predict'
    )

    session.add(blo)
    session.commit()


def eval_stacked_ensemble(stacked_ensemble, session, path):
    """Evaluate stacked ensemble

    Args:
        stacked_ensemble (xcessiv.models.StackedEnsemble)

        session: Valid SQLAlchemy session

        path (str, unicode): Path to project folder

    Returns:
        stacked_ensemble (xcessiv.models.StackedEnsemble)
    """
    try:
        meta_features_list = []
        for base_learner in stacked_ensemble.base_learners:
            mf = np.load(base_learner.meta_features_path(path))
            if len(mf.shape) == 1:
                mf = mf.reshape(-1, 1)
            meta_features_list.append(mf)

        secondary_features = np.concatenate(meta_features_list, axis=1)

        # Get data
        extraction = session.query(models.Extraction).first()
        return_splits_iterable = functions.import_object_from_string_code(
            extraction.meta_feature_generation['source'],
            'return_splits_iterable'
        )
        X, y = extraction.return_train_dataset()

        #  We need to retrieve original order of meta-features
        indices_list = [test_index for train_index, test_index in return_splits_iterable(X, y)]
        indices = np.concatenate(indices_list)
        X, y = X[indices], y[indices]

        est = stacked_ensemble.return_secondary_learner()

        return_splits_iterable_stacked_ensemble = functions.import_object_from_string_code(
            extraction.stacked_ensemble_cv['source'],
            'return_splits_iterable'
        )
        preds = []
        trues_list = []
        for train_index, test_index in return_splits_iterable_stacked_ensemble(secondary_features, y):
            X_train, X_test = secondary_features[train_index], secondary_features[test_index]
            y_train, y_test = y[train_index], y[test_index]
            est = est.fit(X_train, y_train)
            preds.append(
                getattr(est, stacked_ensemble.base_learner_origin.
                        meta_feature_generator)(X_test)
            )
            trues_list.append(y_test)
        preds = np.concatenate(preds, axis=0)
        y_true = np.concatenate(trues_list)

        for key in stacked_ensemble.base_learner_origin.metric_generators:
            metric_generator = functions.import_object_from_string_code(
                stacked_ensemble.base_learner_origin.metric_generators[key],
                'metric_generator'
            )
            stacked_ensemble.individual_score[key] = metric_generator(y_true, preds)

        stacked_ensemble.job_status = 'finished'
        session.add(stacked_ensemble)
        session.commit()
        return stacked_ensemble

    except:
        session.rollback()
        stacked_ensemble.job_status = 'errored'
        stacked_ensemble.description['error_type'] = repr(sys.exc_info()[0])
        stacked_ensemble.description['error_value'] = repr(sys.exc_info()[1])
        stacked_ensemble.description['error_traceback'] = \
            traceback.format_exception(*sys.exc_info())
        session.add(stacked_ensemble)
        session.commit()
        raise


def start_greedy_ensemble_search(automated_run, session, path):
    """Starts an automated ensemble search using greedy forward model selection.

    The steps for this search are adapted from "Ensemble Selection from Libraries of Models" by
    Caruana.

    1. Start with the empty ensemble

    2. Add to the ensemble the model in the library that maximizes the ensemmble's
    performance on the error metric.

    3. Repeat step 2 for a fixed number of iterations or until all models have been used.

    Args:
        automated_run (xcessiv.models.AutomatedRun): Automated run object

        session: Valid SQLAlchemy session

        path (str, unicode): Path to project folder
    """
    module = functions.import_string_code_as_module(automated_run.source)
    assert module.metric_to_optimize in automated_run.base_learner_origin.metric_generators

    best_ensemble = []  # List containing IDs of best performing ensemble for the last round

    secondary_learner = automated_run.base_learner_origin.return_estimator()
    secondary_learner.set_params(**module.secondary_learner_hyperparameters)

    for i in range(module.max_num_base_learners):
        best_score = -float('inf')  # Best metric for this round (not in total!)
        current_ensemble = best_ensemble[:]  # Shallow copy of best ensemble
        for base_learner in session.query(models.BaseLearner).filter_by(job_status='finished').all():
            if base_learner in current_ensemble:  # Don't append when learner is already in
                continue
            current_ensemble.append(base_learner)

            # Check if our "best ensemble" already exists
            existing_ensemble = session.query(models.StackedEnsemble).\
                filter_by(base_learner_origin_id=automated_run.base_learner_origin.id,
                          secondary_learner_hyperparameters=secondary_learner.get_params(),
                          base_learner_ids=sorted([bl.id for bl in current_ensemble])).first()

            if existing_ensemble and existing_ensemble.job_status == 'finished':
                score = existing_ensemble.individual_score[module.metric_to_optimize]

            elif existing_ensemble and existing_ensemble.job_status != 'finished':
                eval_stacked_ensemble(existing_ensemble, session, path)
                score = existing_ensemble.individual_score[module.metric_to_optimize]

            else:
                stacked_ensemble = models.StackedEnsemble(
                    secondary_learner_hyperparameters=secondary_learner.get_params(),
                    base_learners=current_ensemble,
                    base_learner_origin=automated_run.base_learner_origin,
                    job_status='started'
                )
                session.add(stacked_ensemble)
                session.commit()
                eval_stacked_ensemble(stacked_ensemble, session, path)
                score = stacked_ensemble.individual_score[module.metric_to_optimize]

            score = -score if module.invert_metric else score

            if best_score < score:
                best_score = score
                best_ensemble = current_ensemble[:]

            current_ensemble.pop()



from __future__ import absolute_import, print_function, division, unicode_literals
import unittest
import numpy as np
from xcessiv import stacker
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import StratifiedKFold, train_test_split
from sklearn.datasets import load_iris
from sklearn.metrics import accuracy_score


class TestStacker(unittest.TestCase):
    def setUp(self):
        bl1 = RandomForestClassifier(random_state=8)
        bl2 = LogisticRegression()
        bl3 = RandomForestClassifier(max_depth=10, random_state=10)

        meta_est = LogisticRegression()

        skf = StratifiedKFold(random_state=8).split

        self.stacked_ensemble = stacker.XcessivStackedEnsemble(
            [bl1, bl2, bl3],
            ['predict', 'predict_proba', 'predict_proba'],
            meta_est,
            skf
        )

    def test_fit_and_process_using_meta_feature_generator(self):
        X, y = load_iris(return_X_y=True)
        X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=8)

        self.stacked_ensemble.fit(X_train, y_train)

        preds = self.stacked_ensemble._process_using_meta_feature_generator(X_test, 'predict')
        assert np.round(accuracy_score(y_test, preds), 3) == 0.868

        probas = self.stacked_ensemble._process_using_meta_feature_generator(X_test, 'predict_proba')
        preds = np.argmax(probas, axis=1)
        assert np.round(accuracy_score(y_test, preds), 3) == 0.868

    def test_get_params(self):
        self.stacked_ensemble.get_params()

    def test_set_params(self):
        self.stacked_ensemble.set_params(bl0__random_state=20)
        assert self.stacked_ensemble.get_params()['bl0__random_state'] == 20
        assert self.stacked_ensemble.get_params()['bl0'].get_params()['random_state'] == 20

        self.stacked_ensemble.set_params(**{'secondary-learner__C': 1.5})
        assert self.stacked_ensemble.get_params()['secondary-learner__C'] == 1.5
        assert self.stacked_ensemble.get_params()['secondary-learner'].get_params()['C'] == 1.5
